;****************************************************************
;*           HF Synthesizer on AD9834 DDS                       *
;*           Version 0.2.0 25.11.2010                           *
;*                                                              *
;*           Copyright (C) 2010 UR4MCK                          *
;*            This program is FREEWARE                          *
;*                                                              *
;*              AD9834 DDS ROUTINE                              *
;*                                                              *
;*      25.11.2010: second version                              *
;*      01.09.2010: first version                               *
;*                                                              *
;****************************************************************

;---------------------------------------------------------
;		DDS_INIT - Initialize DDS chip
;
;	Input:	none
;	Output:	none
;	Used:	tmp, cnt (all not saved)
;---------------------------------------------------------
dds_init:
		; Switch SPI to Mode 2
		sbi SPCR, CPOL

		; Reset DDS chip (CR = 0x2100)
		cbi DDS_PORT, DDS_FSYNC_PIN	; Start of data
		ldi tmp, 0x21				; Control Register (high byte)
		rcall spi_send
		ldi tmp, 0x00				; Control Register (low byte)
		rcall spi_send
		sbi DDS_PORT, DDS_FSYNC_PIN	; End of data

		; 10 ms pause to let DDS to wakeup
		ldi tmp, 10
		rcall delay_ms

		; Remove reset (CR = 0x2000)
		cbi DDS_PORT, DDS_FSYNC_PIN	; Start of data
		ldi tmp, 0x20				; Control Register (high byte)
		rcall spi_send
		ldi tmp, 0x00				; Control Register (low byte)
		rcall spi_send
		sbi DDS_PORT, DDS_FSYNC_PIN	; End of data

		; Switch SPI back to Mode 0
		cbi SPCR, CPOL

		ret

;-------------------------------------------------------------------------
;		DDS_TUNE - Tune DDS chip;
;	Input: 	aux3...0 (not saved)
;	Output:	none
;	Used:	tmp (not saved)
;-------------------------------------------------------------------------
dds_tune:
		; Copy new frequency value to other FREQ register
		sbrs ctrl, CTRL_DDS_REG
		rjmp dds_tune_1
		ldi tmp, 0b01000000			; Use FREQ0
		cbr ctrl, (1 << CTRL_DDS_REG)
		rjmp dds_tune_2

dds_tune_1:
		ldi tmp, 0b10000000			; Use FREQ1
		sbr ctrl, (1 << CTRL_DDS_REG)

dds_tune_2:
		or aux1, tmp
		or aux3, tmp

		; Switch SPI to Mode 2
		sbi SPCR, CPOL

		; Write to FREQn LSB
		cbi DDS_PORT, DDS_FSYNC_PIN	; Start of data
		mov tmp, aux1				; FREQn LSB (high byte)
		rcall spi_send
		mov tmp, aux0				; FREQn LSB (low byte)
		rcall spi_send

		; Write to FREQn MSB
		mov tmp, aux3				; FREQn MSB (high byte)
		rcall spi_send
		mov tmp, aux2				; FREQn MSB (low byte)
		rcall spi_send
		sbi DDS_PORT, DDS_FSYNC_PIN	; End of data

		; Adjust control register
		cbi DDS_PORT, DDS_FSYNC_PIN	; Start of data
		ldi tmp, 0x28				; CR: FREQ1 selected  (high byte)
		sbrs ctrl, CTRL_DDS_REG
		ldi tmp, 0x20				; CR: FREQ0 selected (high byte)
		rcall spi_send
		ldi tmp, 0x00				; CR (low byte)
		rcall spi_send
		sbi DDS_PORT, DDS_FSYNC_PIN	; End of data

		; Switch SPI back to Mode 0
		cbi SPCR, CPOL
		ret

;---------------------------------------------------------
;		DDS_FREQ2FTW - Calculate DDS FTW from tuning frequency
;
;	Input:	FREQ_TUNING - Tuning frequency (unpacked BCD)
;	Output:	aux0..3 - formatted FTW for DDS (not saved)
;	Used:	tmp, cnt, reg0..7 (all not saved)
;---------------------------------------------------------
dds_freq2ftw:
		; Calculate f6 component
		ldi tmp, low(DDS_FACTOR >> 16)
		mov reg2, tmp
		ldi tmp, low(DDS_FACTOR >> 8)
		mov reg1, tmp
		ldi tmp, low(DDS_FACTOR)
		mov reg0, tmp
		lds tmp, FREQ_TUNING + 7	; Digit 7: 1x.xxx.xxx
		ldi cnt, 10
		mul tmp, cnt				; d7 * 10
		lds tmp, FREQ_TUNING + 6	; Digit 6: x1.xxx.xxx
		add tmp, aux0
		rcall umul8x24				; f6 = (d7..d6) * b6
		movw reg4, aux0
		movw reg6, aux2

		; Calculate f5 component
		ldi tmp, low((DDS_FACTOR / 10) >> 16)
		mov reg2, tmp
		ldi tmp, low((DDS_FACTOR / 10) >> 8)
		mov reg1, tmp
		ldi tmp, low(DDS_FACTOR / 10)
		mov reg0, tmp
		lds tmp, FREQ_TUNING + 5	; Digit 5: xx.1xx.xxx
		rcall umul8x24				; f5 = d5 * b5
		add reg4, aux0
		adc reg5, aux1
		adc reg6, aux2
		adc reg7, aux3

		; Calculate f4 component
		ldi tmp, low((DDS_FACTOR / 100) >> 16)
		mov reg2, tmp
		ldi tmp, low((DDS_FACTOR / 100) >> 8)
		mov reg1, tmp
		ldi tmp, low(DDS_FACTOR / 100)
		mov reg0, tmp
		lds tmp, FREQ_TUNING + 4	; Digit 4: xx.x1x.xxx
		rcall umul8x24				; f4 = d4 * b4
		add reg4, aux0
		adc reg5, aux1
		adc reg6, aux2
		adc reg7, aux3

		; Calculate f3 component
		ldi tmp, low((DDS_FACTOR / 1000) >> 8)
		mov reg1, tmp
		ldi tmp, low(DDS_FACTOR / 1000)
		mov reg0, tmp
		lds tmp, FREQ_TUNING + 3	; Digit 3: xx.xx1.xxx
		rcall umul8x16				; f3 = d3 * b3
		add reg4, aux0
		adc reg5, aux1
		adc reg6, aux2

		; Calculate f2 component
		ldi tmp, low((DDS_FACTOR / 10000) >> 8)
		mov reg1, tmp
		ldi tmp, low(DDS_FACTOR / 10000)
		mov reg0, tmp
		lds tmp, FREQ_TUNING + 2	; Digit 2: xx.xxx.1xx
		rcall umul8x16				; f2 = d2 * b2
		add reg4, aux0
		adc reg5, aux1
		adc reg6, aux2

		; Calculate f1 component
		ldi cnt, low(DDS_FACTOR / 100000)
		lds tmp, FREQ_TUNING + 1	; Digit 1: xx.xxx.x1x
		mul tmp, cnt				; f1 = d1 * b1
		add reg4, aux0
		adc reg5, aux1

		; Calculate f0 component
		ldi cnt, low(DDS_FACTOR / 1000000)
		lds tmp, FREQ_TUNING		; Digit 0: xx.xxx.xx1
		mul tmp, cnt				; f0 = d0 * b0
		add reg4, aux0
		adc reg5, aux1
		; NOTE: now in reg7 - reg4 - calculated FTW value

		; Convert it to DDS specific representation
		mov tmp, reg5
		rol reg5
		rol reg6
		rol reg7
		rol reg5
		rol reg6
		rol reg7
		mov reg5, tmp
		ldi tmp, 0b00111111		; Mask for FREQn select bits
		and reg5, tmp			; Actual FREQn select bits would be added
		and reg7, tmp			; in dds_tune()

		; Move result to its destination
		movw aux0, reg4			; FREG0 register value
		movw aux2, reg6			; FREG1 register value
		ret

;---------------------------------------------------------
;		DDS_UPDATE - Update DDS data & perform tuning
;
;	Input:	DDS tuning data & status
;	Output:	FREQ_TUNING, FREQ_DISPLAY
;	Used:	various registers in subroutines (all not saved)
;---------------------------------------------------------
dds_update:
		sbrs ctrl, CTRL_VFO	; Check VFO type
		rjmp dds_update_vfoa

		; Use VFO_B_FREQ
		ldi XH, high(FREQ_VFO_B)
		ldi XL, low(FREQ_VFO_B)
		rjmp dds_update_rxtx

dds_update_vfoa:
		; Use VFO_A_FREQ
		ldi XH, high(FREQ_VFO_A)
		ldi XL, low(FREQ_VFO_A)

dds_update_rxtx:
		; Load all frequency BCD digits
		ld reg0, X+
		ld reg1, X+
		ld reg2, X+
		ld reg3, X+
		ld reg4, X+
		ld reg5, X+
		ld reg6, X+
		ld reg7, X

		sbrs ctrl, CTRL_RX_TX	; Check RX/TX mode
		rjmp dds_update_rx

		; Use XIT
		ldi XH, high(FREQ_XIT)
		ldi XL, low(FREQ_XIT)
		rjmp dds_update_rit

		; RX mode
dds_update_rx:
		; Use RIT
		ldi XH, high(FREQ_RIT)
		ldi XL, low(FREQ_RIT)

dds_update_rit:
		; F = F0 + RIT(XIT)
		ld tmp, X+
		add reg0, tmp
		ld tmp, X+
		adc reg1, tmp
		ld tmp, X+
		adc reg2, tmp
		ld tmp, X+
		adc reg3, tmp
		ld tmp, X+
		adc reg4, tmp
		ld tmp, X+
		adc reg5, tmp
		ld tmp, X+
		adc reg6, tmp
		ld tmp, X
		adc reg7, tmp

		; Store display frequency
		sts FREQ_DISPLAY, reg0
		sts FREQ_DISPLAY + 1, reg1
		sts FREQ_DISPLAY + 2, reg2
		sts FREQ_DISPLAY + 3, reg3
		sts FREQ_DISPLAY + 4, reg4
		sts FREQ_DISPLAY + 5, reg5
		sts FREQ_DISPLAY + 6, reg6
		sts FREQ_DISPLAY + 7, reg7

		sbrs ctrl, CTRL_RX_TX	; Check RX/TX mode again
		rjmp dds_update_rx_shift

		; Use TX_SHIFT
		ldi XH, high(FREQ_TX_SHIFT)
		ldi XL, low(FREQ_TX_SHIFT)
		rjmp dds_update_shift

		; Use RX_SHIFT
dds_update_rx_shift:
		ldi XH, high(FREQ_RX_SHIFT)
		ldi XL, low(FREQ_RX_SHIFT)

dds_update_shift:
		; F = F + RX(TX)_SHIFT
		ld tmp, X+
		add reg0, tmp
		ld tmp, X+
		adc reg1, tmp
		ld tmp, X+
		adc reg2, tmp
		ld tmp, X+
		adc reg3, tmp
		ld tmp, X+
		adc reg4, tmp
		ld tmp, X+
		adc reg5, tmp
		ld tmp, X+
		adc reg6, tmp
		ld tmp, X
		adc reg7, tmp

		; Store tuning frequency
		sts FREQ_TUNING, reg0
		sts FREQ_TUNING + 1, reg1
		sts FREQ_TUNING + 2, reg2
		sts FREQ_TUNING + 3, reg3
		sts FREQ_TUNING + 4, reg4
		sts FREQ_TUNING + 5, reg5
		sts FREQ_TUNING + 6, reg6
		sts FREQ_TUNING + 7, reg7

		rcall dds_freq2ftw	; Convert frequency to DDS FTW
		rcall dds_tune		; Perform DDS tuning
		ret
